/*
 * serial.c
 *
 *  Created on: 28 lis 2017
 *      Author: dsosnowski
 */


#include "chip.h"
#include "system.h"
#include "serial.h"


struct {
	uint8_t writeBuff[SERIAL_WBUFFER_SIZE];
	volatile uint16_t writeCounter;
	volatile uint8_t sendStatus;
	volatile uint8_t* sendPt;
	volatile uint16_t sendCounter;

	uint8_t readBuff[SERIAL_RBUFFER_SIZE];
	volatile uint16_t readCounter;
	volatile uint8_t receiveStatus;

	uint16_t oneCharTimerTicks;
}serial;


#define SERIAL_TIMER				LPC_TIMER16_0
#define SERIAL_TIMER_PRESCALER		47
#define SERIAL_TIMER_MATCHNUM		0
#define LPC_SERIAL					LPC_USART

uint32_t serialbaudVal[] = {2400, 4800, 9600, 14400, 19200, 28800, 57600, 115200};


void SerialTimerResetSet(LPC_TIMER_T *pTMR);
void SerialTimerBaudCalc(LPC_TIMER_T *pTMR, uint32_t baudrate);


//----------------------------------------------------------------------------------
void SerialInit(uint32_t baudrate)
{
	/* Enable IOCON clock */
	//Chip_Clock_EnablePeriphClock(SYSCTL_CLOCK_IOCON);
    Chip_IOCON_PinMux(LPC_IOCON, IOCON_PIO1_6, IOCON_FUNC1, IOCON_MODE_PULLUP);	//RXD
    Chip_IOCON_PinMux(LPC_IOCON, IOCON_PIO1_7, IOCON_FUNC1, IOCON_MODE_INACT);	//TXD

    Chip_UART_Init(LPC_SERIAL);
	SerialSetBaud(LPC_SERIAL, baudrate);
	//Chip_UART_SetupFIFOS(LPC_USART, (UART_FCR_FIFO_EN | UART_FCR_TRG_LEV0));
	Chip_UART_TXEnable(LPC_SERIAL);

	SerialTimerInit();
	SerialTimerBaudCalc(SERIAL_TIMER, baudrate);

	NVIC_ClearPendingIRQ(TIMER_16_0_IRQn);
	NVIC_EnableIRQ(TIMER_16_0_IRQn);

	/* Enable timer interrupt */
	Chip_UART_IntEnable(LPC_USART, UART_IER_RBRINT);
	NVIC_ClearPendingIRQ(UART0_IRQn);
	NVIC_EnableIRQ(UART0_IRQn);
}

//----------------------------------------------------------------------------------
void SerialSetBaud(LPC_USART_T *pUART, uint32_t baudrate)
{
	uint32_t div, divh, divl;

	div = MY_MAIN_CLOCK / (baudrate * 16);
	/* High and low halves of the divider */
	divh = div / 256;
	divl = div - (divh * 256);

	Chip_UART_EnableDivisorAccess(pUART);
	Chip_UART_SetDivisorLatches(pUART, divl, divh);
	Chip_UART_DisableDivisorAccess(pUART);

	SerialTimerBaudCalc(SERIAL_TIMER, baudrate);
}

//-------------------------------------------------------------------
void SerialTimerInit(void) {
	Chip_TIMER_Init(SERIAL_TIMER);
	SerialTimerResetSet(SERIAL_TIMER);
	Chip_TIMER_PrescaleSet(SERIAL_TIMER, SERIAL_TIMER_PRESCALER);

	Chip_TIMER_SetMatch(SERIAL_TIMER, SERIAL_TIMER_MATCHNUM, 4);
	Chip_TIMER_StopOnMatchEnable(SERIAL_TIMER, SERIAL_TIMER_MATCHNUM);
	Chip_TIMER_MatchEnableInt(SERIAL_TIMER, SERIAL_TIMER_MATCHNUM);
}

//-------------------------------------------------------------------
void SerialTimerResetSet(LPC_TIMER_T *pTMR) {
	/* Disable timer, set terminal count to non-0 */
	pTMR->TCR = 0;
	pTMR->TC = 1;
	/* Reset timer counter */
	pTMR->TCR = TIMER_RESET;
	/* Wait for terminal count to clear */
	while (pTMR->TC != 0) {}
	pTMR->TCR = 0;
}

//-------------------------------------------------------------------
void SerialTimerBaudCalc(LPC_TIMER_T *pTMR, uint32_t baudrate) {
	uint32_t result;
	result = (MY_MAIN_CLOCK / SERIAL_TIMER_PRESCALER);
	result *= 10;
	result /= baudrate;
	result &= 0xFFFF;
	serial.oneCharTimerTicks = result;
}

//-------------------------------------------------------------------
void SerialTimerStop(void) {
	SerialTimerResetSet(SERIAL_TIMER);
}

//-------------------------------------------------------------------
void SerialTimerStart(uint8_t charsDelay) {
	SERIAL_TIMER->TC = 0;
	Chip_TIMER_SetMatch(SERIAL_TIMER, SERIAL_TIMER_MATCHNUM, (charsDelay * serial.oneCharTimerTicks));
	Chip_TIMER_Enable(SERIAL_TIMER);
}


//-------------------------------------------------------------------
void TIMER16_0_IRQHandler(void){
	Chip_TIMER_ClearMatch(SERIAL_TIMER, SERIAL_TIMER_MATCHNUM);
	serial.sendStatus = SERIAL_READY;
	if (serial.readCounter > SERIAL_MIN_FRAME_LEN){
		serial.receiveStatus = SERIAL_READY;
	} else {
		SerialRBufferClear();
	}
}

//----------------------------------------------------------------------------------
void UART_IRQHandler(void){
	uint32_t statusReg = Chip_UART_ReadLineStatus(LPC_SERIAL);
	uint8_t readData;

	if ((statusReg & UART_LSR_THRE) > 0){
		if (serial.sendCounter < serial.writeCounter){
			Chip_UART_SendByte(LPC_SERIAL, *serial.sendPt);
			serial.sendCounter++;
			serial.sendPt++;
		} else {
			if (serial.sendStatus == SERIAL_BUSY){
				Chip_UART_IntDisable(LPC_SERIAL, UART_IER_THREINT);
				serial.sendStatus = SERIAL_ENDING;
				SerialTimerStart(5);
			}
		}
	}

	if ((statusReg & UART_LSR_RDR) > 0){
		readData = Chip_UART_ReadByte(LPC_SERIAL);
		//SerialTimerStart();
		if ((serial.receiveStatus < SERIAL_READY) && (serial.sendStatus == SERIAL_IDLE)){
			serial.receiveStatus = SERIAL_BUSY;

			SerialTimerStart(4);
			if (readData != SERIAL_ENTER_CHAR) {
				SerialTimerStop();
				//serial.receiveStatus = SERIAL_READY;
			}

			if (serial.readCounter < (SERIAL_RBUFFER_SIZE - 1)){
				serial.readBuff[serial.readCounter] = readData;
				serial.readCounter++;
				serial.readBuff[serial.readCounter] = 0;
			}
		}
	}
}

//-------------------------------------------
uint8_t SerialWBufferAdd(uint8_t *src){
	char status = 0;
	if (serial.sendStatus > SERIAL_IDLE) status = 3;
	while (status == 0) {
		if ((*src) && (src)) {
			if (serial.writeCounter < (SERIAL_WBUFFER_SIZE - 1)) {
				serial.writeBuff[serial.writeCounter] = *src;
				serial.writeCounter++;
				src++;
			} else status = 2;
		} else status = 1;

		serial.writeBuff[serial.writeCounter] = 0;
	}
	return status;
}

//-------------------------------------------
uint8_t SerialWBufferAddLen(uint8_t *src, uint8_t len){
	char status = 0;
	if (serial.sendStatus > SERIAL_IDLE) status = 3;
	while (status == 0) {
		if (len) {
			if (serial.writeCounter < (SERIAL_WBUFFER_SIZE - 1)) {
				serial.writeBuff[serial.writeCounter] = *src;
				serial.writeCounter++;
				src++;
				len--;
			} else status = 2;
		} else status = 1;

		serial.writeBuff[serial.writeCounter] = 0;
	}
	return status;
}

//-------------------------------------------
void SerialWBufferSend(void){
	if (serial.writeCounter > 0){
		if ((serial.sendStatus == SERIAL_IDLE) && (serial.receiveStatus != SERIAL_BUSY)){
			serial.sendStatus = SERIAL_BUSY;
			serial.sendCounter = 0;
			serial.sendPt = &serial.writeBuff[0];

			//while (!(LPC_USART->LSR & UART_LSR_THRE)){};
			Chip_UART_SendByte(LPC_USART, *serial.sendPt);
			serial.sendCounter++;
			serial.sendPt++;

			Chip_UART_IntEnable(LPC_SERIAL, UART_IER_THREINT);
		}
	}
}

//----------------------------------------------------
uint8_t SerialWBufferSendCompleteCheck(void){
	if (serial.sendStatus == SERIAL_READY) return 1;
	return 0;
}

//-------------------------------------------
void SerialWBufferNew(void){
	if (serial.sendStatus >= SERIAL_READY){
		serial.writeCounter = 0;
		serial.writeBuff[0] = 0;
		serial.sendPt = &serial.writeBuff[0];
		serial.sendStatus = SERIAL_IDLE;
	}
}

//----------------------------------------------------
uint8_t SerialRBufferCheck(void){
	return serial.receiveStatus;
}

//----------------------------------------------------
uint8_t SerialRBufferCounter(void){
	return serial.readCounter;
}

//----------------------------------------------------
void SerialRBufferClear(void){
	serial.receiveStatus = SERIAL_IDLE;
	serial.readCounter = 0;
	serial.readBuff[0] = 0;
}

//----------------------------------------------------
uint8_t SerialRBufferChar(uint8_t index){
	return serial.readBuff[index];
}

//----------------------------------------------------
uint8_t *SerialRBufferSrc(uint8_t index){
	return &serial.readBuff[index];
}
