/*
 * drvST7735.c
 *
 *  Created on: 05.02.2018
 *      Author: Procesorowiec
 */


#include "chip.h"
#include "system.h"
#include "mydelay.h"
#include "myspi.h"
#include "drvST7735.h"

//================================================================
static void ST7735DelayMs(uint16_t ms){
	MyDelayMs(ms);
}

//----------------------------------------------------------------
static void ST7735HardReset(void){
	MySpiHResetSet();
	ST7735DelayMs(200);
	MySpiHResetRelase();
	ST7735DelayMs(200);
}

//----------------------------------------------------------------
static uint8_t ST7735ReadyToSend(void){
	return MySpiReadyToSend();
}

//----------------------------------------------------------------
static void ST7735SendByte(uint8_t data){
	MySpiSendByte(data);
}

//----------------------------------------------------------------
static uint8_t ST7735SendComplete(void){
	return MySpiSendComplete();
}

//----------------------------------------------------------------
static void ST7735CommandModeSet(void){
	MySpiCommandModeSet();
}

//----------------------------------------------------------------
static void ST7735DataModeSet(void){
	MySpiDataModeSet();
}


//================================================================
#define ST7735_TFTWIDTH  128
#define ST7735_TFTHEIGHT 160

#define ST7735_NOP     0x00
#define ST7735_SWRESET 0x01
#define ST7735_RDDID   0x04
#define ST7735_RDDST   0x09

#define ST7735_SLPIN   0x10
#define ST7735_SLPOUT  0x11
#define ST7735_PTLON   0x12
#define ST7735_NORON   0x13

#define ST7735_INVOFF  0x20
#define ST7735_INVON   0x21
#define ST7735_DISPOFF 0x28
#define ST7735_DISPON  0x29
#define ST7735_CASET   0x2A
#define ST7735_RASET   0x2B
#define ST7735_RAMWR   0x2C
#define ST7735_RAMRD   0x2E

#define ST7735_PTLAR   0x30
#define ST7735_COLMOD  0x3A
#define ST7735_MADCTL  0x36

#define ST7735_FRMCTR1 0xB1
#define ST7735_FRMCTR2 0xB2
#define ST7735_FRMCTR3 0xB3
#define ST7735_INVCTR  0xB4
#define ST7735_DISSET5 0xB6

#define ST7735_PWCTR1  0xC0
#define ST7735_PWCTR2  0xC1
#define ST7735_PWCTR3  0xC2
#define ST7735_PWCTR4  0xC3
#define ST7735_PWCTR5  0xC4
#define ST7735_VMCTR1  0xC5

#define ST7735_RDID1   0xDA
#define ST7735_RDID2   0xDB
#define ST7735_RDID3   0xDC
#define ST7735_RDID4   0xDD

#define ST7735_PWCTR6  0xFC

#define ST7735_GMCTRP1 0xE0
#define ST7735_GMCTRN1 0xE1

//----------------------------------------------------------------
//----------------------------------------------------------------
//----------------------------------------------------------------
// Rather than a bazillion writecommand() and writedata() calls, screen
// initialization commands and arguments are organized in these tables
// stored in ROM.  The table may look bulky, but that's mostly the
// formatting -- storage-wise this is hundreds of bytes more compact
// than the equivalent code.  Companion function follows.
#define SEQ_DELAY 0x80
static const uint8_t
  Bseq[] = {                  // Initialization commands for 7735B screens
    18,                       // 18 commands in list:
    ST7735_SWRESET,   SEQ_DELAY,  //  1: Software reset, no args, w/delay
      50,                     //     50 ms delay
    ST7735_SLPOUT ,   SEQ_DELAY,  //  2: Out of sleep mode, no args, w/delay
      255,                    //     255 = 500 ms delay
    ST7735_COLMOD , 1+SEQ_DELAY,  //  3: Set color mode, 1 arg + delay:
      0x05,                   //     16-bit color
      10,                     //     10 ms delay
    ST7735_FRMCTR1, 3+SEQ_DELAY,  //  4: Frame rate control, 3 args + delay:
      0x00,                   //     fastest refresh
      0x06,                   //     6 lines front porch
      0x03,                   //     3 lines back porch
      10,                     //     10 ms delay
    ST7735_MADCTL , 1      ,  //  5: Memory access ctrl (directions), 1 arg:
      0x08,                   //     Row addr/col addr, bottom to top refresh
    ST7735_DISSET5, 2      ,  //  6: Display settings #5, 2 args, no delay:
      0x15,                   //     1 clk cycle nonoverlap, 2 cycle gate
                              //     rise, 3 cycle osc equalize
      0x02,                   //     Fix on VTL
    ST7735_INVCTR , 1      ,  //  7: Display inversion control, 1 arg:
      0x0,                    //     Line inversion
    ST7735_PWCTR1 , 2+SEQ_DELAY,  //  8: Power control, 2 args + delay:
      0x02,                   //     GVDD = 4.7V
      0x70,                   //     1.0uA
      10,                     //     10 ms delay
    ST7735_PWCTR2 , 1      ,  //  9: Power control, 1 arg, no delay:
      0x05,                   //     VGH = 14.7V, VGL = -7.35V
    ST7735_PWCTR3 , 2      ,  // 10: Power control, 2 args, no delay:
      0x01,                   //     Opamp current small
      0x02,                   //     Boost frequency
    ST7735_VMCTR1 , 2+SEQ_DELAY,  // 11: Power control, 2 args + delay:
      0x3C,                   //     VCOMH = 4V
      0x38,                   //     VCOML = -1.1V
      10,                     //     10 ms delay
    ST7735_PWCTR6 , 2      ,  // 12: Power control, 2 args, no delay:
      0x11, 0x15,
    ST7735_GMCTRP1,16      ,  // 13: Magical unicorn dust, 16 args, no delay:
      0x09, 0x16, 0x09, 0x20, //     (seriously though, not sure what
      0x21, 0x1B, 0x13, 0x19, //      these config values represent)
      0x17, 0x15, 0x1E, 0x2B,
      0x04, 0x05, 0x02, 0x0E,
    ST7735_GMCTRN1,16+SEQ_DELAY,  // 14: Sparkles and rainbows, 16 args + delay:
      0x0B, 0x14, 0x08, 0x1E, //     (ditto)
      0x22, 0x1D, 0x18, 0x1E,
      0x1B, 0x1A, 0x24, 0x2B,
      0x06, 0x06, 0x02, 0x0F,
      10,                     //     10 ms delay
    ST7735_CASET  , 4      ,  // 15: Column addr set, 4 args, no delay:
      0x00, 0x02,             //     XSTART = 2
      0x00, 0x81,             //     XEND = 129
    ST7735_RASET  , 4      ,  // 16: Row addr set, 4 args, no delay:
      0x00, 0x02,             //     XSTART = 1
      0x00, 0x81,             //     XEND = 160
    ST7735_NORON  ,   SEQ_DELAY,  // 17: Normal display on, no args, w/delay
      10,                     //     10 ms delay
    ST7735_DISPON ,   SEQ_DELAY,  // 18: Main screen turn on, no args, w/delay
      255 };                  //     255 = 500 ms delay
static const uint8_t
  Rseq1[] = {                 // Init for 7735R, part 1 (red or green tab)
    15,                       // 15 commands in list:
    ST7735_SWRESET,   SEQ_DELAY,  //  1: Software reset, 0 args, w/delay
      150,                    //     150 ms delay
    ST7735_SLPOUT ,   SEQ_DELAY,  //  2: Out of sleep mode, 0 args, w/delay
      255,                    //     500 ms delay
    ST7735_FRMCTR1, 3      ,  //  3: Frame rate ctrl - normal mode, 3 args:
      0x01, 0x2C, 0x2D,       //     Rate = fosc/(1x2+40) * (LINE+2C+2D)
    ST7735_FRMCTR2, 3      ,  //  4: Frame rate control - idle mode, 3 args:
      0x01, 0x2C, 0x2D,       //     Rate = fosc/(1x2+40) * (LINE+2C+2D)
    ST7735_FRMCTR3, 6      ,  //  5: Frame rate ctrl - partial mode, 6 args:
      0x01, 0x2C, 0x2D,       //     Dot inversion mode
      0x01, 0x2C, 0x2D,       //     Line inversion mode
    ST7735_INVCTR , 1      ,  //  6: Display inversion ctrl, 1 arg, no delay:
      0x07,                   //     No inversion
    ST7735_PWCTR1 , 3      ,  //  7: Power control, 3 args, no delay:
      0xA2,
      0x02,                   //     -4.6V
      0x84,                   //     AUTO mode
    ST7735_PWCTR2 , 1      ,  //  8: Power control, 1 arg, no delay:
      0xC5,                   //     VGH25 = 2.4C VGSEL = -10 VGH = 3 * AVDD
    ST7735_PWCTR3 , 2      ,  //  9: Power control, 2 args, no delay:
      0x0A,                   //     Opamp current small
      0x00,                   //     Boost frequency
    ST7735_PWCTR4 , 2      ,  // 10: Power control, 2 args, no delay:
      0x8A,                   //     BCLK/2, Opamp current small & Medium low
      0x2A,
    ST7735_PWCTR5 , 2      ,  // 11: Power control, 2 args, no delay:
      0x8A, 0xEE,
    ST7735_VMCTR1 , 1      ,  // 12: Power control, 1 arg, no delay:
      0x0E,
    ST7735_INVOFF , 0      ,  // 13: Don't invert display, no args, no delay
    ST7735_MADCTL , 1      ,  // 14: Memory access control (directions), 1 arg:
      0xC8,                   //     row addr/col addr, bottom to top refresh
    ST7735_COLMOD , 1      ,  // 15: set color mode, 1 arg, no delay:
      0x05 };                 //     16-bit color
static const uint8_t
  Rseq2green[] = {            // Init for 7735R, part 2 (green tab only)
    2,                        //  2 commands in list:
    ST7735_CASET  , 4      ,  //  1: Column addr set, 4 args, no delay:
      0x00, 0x02,             //     XSTART = 0
      0x00, 0x7F+0x02,        //     XEND = 127
    ST7735_RASET  , 4      ,  //  2: Row addr set, 4 args, no delay:
      0x00, 0x01,             //     XSTART = 0
      0x00, 0x9F+0x01 };      //     XEND = 159
static const uint8_t
  Rseq2red[] = {              // Init for 7735R, part 2 (red tab only)
    2,                        //  2 commands in list:
    ST7735_CASET  , 4      ,  //  1: Column addr set, 4 args, no delay:
      0x00, 0x00,             //     XSTART = 0
      0x00, 0x7F,             //     XEND = 127
    ST7735_RASET  , 4      ,  //  2: Row addr set, 4 args, no delay:
      0x00, 0x00,             //     XSTART = 0
      0x00, 0x9F };           //     XEND = 159

static const uint8_t
  Rseq2green144[] = {              // Init for 7735R, part 2 (green 1.44 tab)
    2,                        //  2 commands in list:
    ST7735_CASET  , 4      ,  //  1: Column addr set, 4 args, no delay:
      0x00, 0x00,             //     XSTART = 0
      0x00, 0x7F,             //     XEND = 127
    ST7735_RASET  , 4      ,  //  2: Row addr set, 4 args, no delay:
      0x00, 0x00,             //     XSTART = 0
      0x00, 0x7F };           //     XEND = 127

static const uint8_t
  Rseq2green160x80[] = {              // Init for 7735R, part 2 (mini 160x80)
    2,                        //  2 commands in list:
    ST7735_CASET  , 4      ,  //  1: Column addr set, 4 args, no delay:
      0x00, 0x00,             //     XSTART = 0
      0x00, 0x7F,             //     XEND = 79
    ST7735_RASET  , 4      ,  //  2: Row addr set, 4 args, no delay:
      0x00, 0x00,             //     XSTART = 0
      0x00, 0x9F };           //     XEND = 159
	  
static const uint8_t
  Rseq3[] = {                 // Init for 7735R, part 3 (red or green tab)
    4,                        //  4 commands in list:
    ST7735_GMCTRP1, 16      , //  1: Magical unicorn dust, 16 args, no delay:
      0x02, 0x1c, 0x07, 0x12,
      0x37, 0x32, 0x29, 0x2d,
      0x29, 0x25, 0x2B, 0x39,
      0x00, 0x01, 0x03, 0x10,
    ST7735_GMCTRN1, 16      , //  2: Sparkles and rainbows, 16 args, no delay:
      0x03, 0x1d, 0x07, 0x06,
      0x2E, 0x2C, 0x29, 0x2D,
      0x2E, 0x2E, 0x37, 0x3F,
      0x00, 0x00, 0x02, 0x10,
    ST7735_NORON  ,    SEQ_DELAY, //  3: Normal display on, no args, w/delay
      10,                     //     10 ms delay
    ST7735_DISPON ,    SEQ_DELAY, //  4: Main screen turn on, no args w/delay
      100 };				  //     100 ms delay

//----------------------------------------------------------------
void ST7735Command(uint8_t command){
	while (ST7735SendComplete());
	ST7735CommandModeSet();
	ST7735SendByte(command);
}

//----------------------------------------------------------------
void ST7735Data(uint8_t *data, uint16_t len){
	while (ST7735SendComplete());
	ST7735DataModeSet();

	while (len > 0){
		if (ST7735ReadyToSend() > 0) {
			ST7735SendByte(*data);
			data++;
			len--;
		}
	}
}

//----------------------------------------------------------------
void ST7735Sequence(uint8_t *offset){
	uint8_t numCommands;
	uint8_t numArgs;
	uint16_t ms;

	numCommands = *offset; offset++;

	while (numCommands){
		ST7735Command(*offset); offset++;
		numArgs = *offset; offset++;
		ms = (numArgs & SEQ_DELAY);

		numArgs &= ~(SEQ_DELAY);
		ST7735Data(offset, numArgs);

		offset += numArgs;

		if (ms > 0){
			ms = *offset; offset++;
			if (ms == 255) ms = 500;
			while (ST7735SendComplete());
			ST7735DelayMs(ms);
		}
		numCommands--;
	}
}

//----------------------------------------------------------------
//----------------------------------------------------------------
//----------------------------------------------------------------
#define INITR_GREENTAB   	0x0
#define INITR_REDTAB     	0x1
#define INITR_BLACKTAB   	0x2

#define INITR_18GREENTAB    INITR_GREENTAB
#define INITR_18REDTAB      INITR_REDTAB
#define INITR_18BLACKTAB    INITR_BLACKTAB
#define INITR_144GREENTAB   0x1
#define INITR_MINI160x80    0x4

//----------------------------------------------------------------
struct {
	uint8_t tabcolor;
	uint8_t colstart;
	uint8_t rowstart;
    uint8_t height;
    uint8_t width;
	uint8_t xstart;
	uint8_t ystart;
} drv;

//----------------------------------------------------------------
void ST7735InitB(void){
	ST7735Sequence(Bseq);
}

//----------------------------------------------------------------
void ST7735InitR(uint8_t options) {
	ST7735Sequence(Rseq1);
  if(options == INITR_GREENTAB) {
    ST7735Sequence(Rseq2green);
    drv.colstart = 2;
    drv.rowstart = 1;
  } else if(options == INITR_144GREENTAB) {
    //_height = ST7735_TFTHEIGHT_128;
    //_width = ST7735_TFTWIDTH_128;
    ST7735Sequence(Rseq2green144);
    drv.colstart = 2;
    drv.rowstart = 3;
  } else if(options == INITR_MINI160x80) {
    //_height = ST7735_TFTHEIGHT_160;
    //_width = ST7735_TFTWIDTH_80;
    ST7735Sequence(Rseq2green160x80);
    drv.colstart = 24;
    drv.rowstart = 0;
  } else {
    ST7735Sequence(Rseq2red);
	drv.colstart = 0;
    drv.rowstart = 0;
  }
  ST7735Sequence(Rseq3);

  uint8_t data = 0xC0;
  // if black, change MADCTL color filter
  if ((options == INITR_BLACKTAB) || (options == INITR_MINI160x80)) {
    ST7735Command(ST7735_MADCTL);
    ST7735Data(&data, 1);
  }

  drv.tabcolor = options;
}

//----------------------------------------------------------------
//----------------------------------------------------------------
//----------------------------------------------------------------

#define MADCTL_MY  0x80
#define MADCTL_MX  0x40
#define MADCTL_MV  0x20
#define MADCTL_ML  0x10
#define MADCTL_RGB 0x00
#define MADCTL_BGR 0x08
#define MADCTL_MH  0x04

void ST7735RotationSet(uint8_t m) {
  uint8_t rotation = (m % 4);
  uint8_t data;
  
  ST7735Command(ST7735_MADCTL);

  switch (rotation) {
   case 0:
     if ((drv.tabcolor == INITR_BLACKTAB) || (drv.tabcolor == INITR_MINI160x80)) {
       data = (MADCTL_MX | MADCTL_MY | MADCTL_RGB);
     } else {
       data = (MADCTL_MX | MADCTL_MY | MADCTL_BGR);
     }
	 ST7735Data(&data, 1);

     if (drv.tabcolor == INITR_144GREENTAB) {
       drv.height = 128;
       drv.width  = 128;
     } else if (drv.tabcolor == INITR_MINI160x80)  {
       drv.height = 160;
       drv.width = 80;
     } else {
       drv.height = 160;
       drv.width  = 128;
     }
     drv.xstart = drv.colstart;
     drv.ystart = drv.rowstart;
     break;
   case 1:
     if ((drv.tabcolor == INITR_BLACKTAB) || (drv.tabcolor == INITR_MINI160x80)) {
       data = (MADCTL_MY | MADCTL_MV | MADCTL_RGB);
     } else {
       data = (MADCTL_MY | MADCTL_MV | MADCTL_BGR);
     }
	 ST7735Data(&data, 1);

     if (drv.tabcolor == INITR_144GREENTAB)  {
       drv.width = 128;
       drv.height = 128;
     } else if (drv.tabcolor == INITR_MINI160x80)  {
       drv.width = 160;
       drv.height = 80;
     } else {
       drv.width = 160;
       drv.height = 128;
     }
     drv.ystart = drv.colstart;
     drv.xstart = drv.rowstart;
     break;
  case 2:
     if ((drv.tabcolor == INITR_BLACKTAB) || (drv.tabcolor == INITR_MINI160x80)) {
       data = (MADCTL_RGB);
     } else {
       data = (MADCTL_BGR);
     }
	 ST7735Data(&data, 1);

     if (drv.tabcolor == INITR_144GREENTAB) {
       drv.height = 128;
       drv.width  = 128;
     } else if (drv.tabcolor == INITR_MINI160x80)  {
       drv.height = 160;
       drv.width = 80;
     } else {
       drv.height = 160;
       drv.width  = 128;
     }
     drv.xstart = drv.colstart;
     drv.ystart = drv.rowstart;
     break;
   case 3:
     if ((drv.tabcolor == INITR_BLACKTAB) || (drv.tabcolor == INITR_MINI160x80)) {
       data = (MADCTL_MX | MADCTL_MV | MADCTL_RGB);
     } else {
       data = (MADCTL_MX | MADCTL_MV | MADCTL_BGR);
     }
	 ST7735Data(&data, 1);

     if (drv.tabcolor == INITR_144GREENTAB)  {
       drv.width = 128;
       drv.height = 128;
     } else if (drv.tabcolor == INITR_MINI160x80)  {
       drv.width = 160;
       drv.height = 80;
     } else {
       drv.width = 160;
       drv.height = 128;
     }
     drv.ystart = drv.colstart;
     drv.xstart = drv.rowstart;
     break;
  }
}

//----------------------------------------------------------------
void ST7735Init(uint8_t rotation){
	ST7735HardReset();
	ST7735InitR(INITR_BLACKTAB);
	ST7735RotationSet(rotation); 		
}

//----------------------------------------------------------------
void DrvWindowSet(uint8_t x1, uint8_t y1, uint8_t x2, uint8_t y2) {
  uint8_t data[4];
	
  ST7735Command(ST7735_CASET); // Column addr set
  data[0] = 0;
  data[1] = (x1 + drv.xstart);
  data[2] = 0;
  data[3] = (x2 + drv.xstart);
  ST7735Data(data, 4);

  ST7735Command(ST7735_RASET); // Row addr set
  data[0] = 0;
  data[1] = (y1 + drv.ystart);
  data[2] = 0;
  data[3] = (y2 + drv.ystart);
  ST7735Data(data, 4);

  ST7735Command(ST7735_RAMWR); // write to RAM

  while (ST7735SendComplete());
  ST7735DataModeSet();
}

//----------------------------------------------------------------
void DrvPixelFill(uint8_t *colorval){
	while (ST7735ReadyToSend() == 0);
	ST7735SendByte(*(colorval + 1));
	while (ST7735ReadyToSend() == 0);
	ST7735SendByte(*colorval);
}

//----------------------------------------------------------------
void DrvDisplayInvert(uint8_t state) {
  ST7735Command(state ? ST7735_INVON : ST7735_INVOFF);
}

//----------------------------------------------------------------
uint8_t DrvWidthGet(void){
	return drv.width;
}

//----------------------------------------------------------------
uint8_t DrvHeightGet(void){
	return drv.height;
}
